using System;
using System.Collections;
using PolCommon;
using System.Web;
using System.Text;
using System.IO;
using System.Runtime.Serialization.Formatters.Soap;

namespace PolClient
{
	/// <p>Proxy handling the finding of the most appropriate Web service instance
	/// and the load balancing.</p>
	/// <p>Copyright (c) 2002 Alexis Grandemange<br/>
	/// Mail: alexis.grandemange@pagebox.net</p>
	/// <pre>This program is free software; you can redistribute it and/or
	/// modify it under the terms of the GNU Lesser General Public
	/// License as published by the Free Software Foundation; version
	/// 2.1 of the License.
	/// This library is distributed in the hope that it will be useful,
	/// but WITHOUT ANY WARRANTY; without even the implied warranty of
	/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	/// GNU Lesser General Public License for more details.
	/// A copy of the GNU Lesser General Public License lesser.txt should be
	/// included in the distribution.</pre>
	public class SmartProxy
	{
		/// <summary>
		/// Webservice Hashtable
		/// Key: WsKey
		/// Value: ArrayList
		/// </summary>
		private static Hashtable wsHt = new Hashtable();
		/// <summary>
		/// Candidate ArrayList
		/// Contains PolarisKey objects
		/// </summary>
		private ArrayList csl;
		/// <summary>
		/// HTTP request
		/// </summary>
		HttpRequest request;

		/// <summary>
		/// Constructor. Populates Hashtable and ArrayList.
		/// </summary>
		/// <param name="repositoryURL">URL of the repository of the Web service archive</param>
		/// <param name="wsArchive">Web service archive</param>
		/// <param name="wsName">Web service</param>
		/// <param name="request">HTTP request object of the source</param>
		public SmartProxy(string repositoryURL, string wsArchive, string wsName, HttpRequest request) 
		{
			this.request = request;
			WsKey wsk = new WsKey(repositoryURL, wsArchive, wsName);
			lock(typeof(SmartProxy)) 
			{
				if (!wsHt.Contains(wsk)) 
				{
					Candidate[] c = GetCandidate(repositoryURL, wsArchive, wsName);
					ArrayList al = new ArrayList();
					for (int i = 0; i < c.Length; ++i) 
					{
						if (c[i].keyRange == null)
							continue;
						PolarisKey pk = new PolarisKey(c[i].keyRange, c[i].url, wsArchive, wsName);
						al.Add(pk);
					}
					wsHt.Add(wsk, al);
				}
			}
			csl = (ArrayList)wsHt[wsk];
		}

		/// <summary>
		/// 1. Builds a ArrayList of possible instances
		/// 2. Choose randomly one of them
		/// </summary>
		/// <param name="userNb">User number</param>
		/// <param name="location">Client logical location</param>
		/// <returns>Web service URL</returns>
		public string GetUrl(int userNb, string location) 
		{
			IEnumerator ide = csl.GetEnumerator();
			ArrayList al = new ArrayList();
			while(ide.MoveNext()) 
			{
				PolarisKey pk = (PolarisKey)ide.Current;
				if (!pk.location.Equals(location))
					continue;
				if ((pk.userLow < userNb) && (pk.userHigh > userNb)) 
					al.Add(pk.url);
			}
			if (al.Count == 0)
				return null;
			Random r = new Random((int)DateTime.Now.Ticks);
			return (string)al[r.Next(al.Count)];
		}

		/// <summary>
		/// Create an ActiveNaming proxy and calls its GetCandidate method.
		/// To retrieve the control PageBox,
		/// assumes that PolClient is installed in the PageBox directory
		/// </summary>
		/// <param name="repositoryURL">URL of the repository of the Web service archive</param>
		/// <param name="wsArchive">Web service archive</param>
		/// <param name="wsName">Web service</param>
		/// <returns>Candidate array</returns>
		private Candidate[] GetCandidate(string repositoryURL, string wsArchive, string wsName) 
		{
			string lpath = request.ServerVariables["PATH_INFO"];
			string host = request.ServerVariables["HTTP_HOST"];
			string port = request.ServerVariables["SERVER_PORT"];
			int pos = lpath.LastIndexOf("/");
			lpath = lpath.Substring(0, pos);
			pos = lpath.LastIndexOf("/");
			lpath = lpath.Substring(0, pos + 1) + "ActiveNaming.asmx";
			string url;
			if (port.Equals("80"))
				url = "http://" + host + lpath;
			else
				url = "http://" + host + ":" + port + lpath;
			// url = "http://localhost/myPageBox/ActiveNaming.asmx";
			ActiveNaming an = new ActiveNaming(url);
			try 
			{
				an.Discover();
				return an.GetCandidate(repositoryURL, wsArchive, wsName);
			}
			catch(Exception e) 
			{
			}
			return null;
		}
	}
	
	/// <summary>
	/// Key of the wsHt Hashtable of SmartProxy
	/// </summary>
	public class WsKey 
	{
		/// <summary>
		/// URL of the repository of the Web service archive
		/// </summary>
		string repositoryURL;
		/// <summary>
		/// Web service archive
		/// </summary>
		string wsArchive;
		/// <summary>
		/// Web service
		/// </summary>
		string wsName;
		/// <summary>
		/// Hash code of the contenation of repositoryURL, wsArchive and wsName.
		/// </summary>
		private int hashCode = 0;

		/// <summary>
		/// Constructor. Sets member variables.
		/// </summary>
		/// <param name="repositoryURL">URL of the repository of the Web service archive</param>
		/// <param name="wsArchive">Web service archive</param>
		/// <param name="wsName">Web service</param>
		public WsKey(string repositoryURL, string wsArchive, string wsName) 
		{
			this.repositoryURL = repositoryURL;
			this.wsArchive = wsArchive;
			this.wsName = wsName;
		}

		/// <summary>
		/// Default constructor.
		/// </summary>
		public WsKey() 
		{
		}

		/// <summary>
		/// Overriden GetHashCode - Needed because WsKey is the key of wsHt
		/// </summary>
		/// <returns>hashcode of the concatenated fields</returns>
		public override int GetHashCode() 
		{
			if (hashCode == 0) 
			{
				string ran = repositoryURL + ";" + wsArchive + ";" + wsName;
				hashCode = ran.GetHashCode();
			}
			return hashCode;
		}

		/// <summary>
		/// Overriden Equals - Needed because WsKey is the key of wsHt
		/// </summary>
		/// <param name="wsk">Other object</param>
		/// <returns>Return true if repositoryURL, wsArchive and wsName have the same value</returns>
		public override bool Equals(object wsk) 
		{
			if (!(wsk is WsKey))
				return false;
			WsKey wsk2 = (WsKey)wsk;
			if (wsk2.repositoryURL.Equals(repositoryURL) && wsk2.wsArchive.Equals(wsArchive) &&
				wsk2.wsName.Equals(wsName))
				return true;
			else
				return false;
		}
	}

	/// <summary>
	/// KeyRange + URL
	/// </summary>
	public class PolarisKey
	{
		/// <summary>
		/// Smallest user number managed by this instance
		/// </summary>
		public int userLow;
		/// <summary>
		/// Highest user number managed by this instance
		/// </summary>
		public int userHigh;
		/// <summary>
		/// Instance location
		/// </summary>
		public string location;
		/// <summary>
		/// Web service URL
		/// </summary>
		public string url;

		/// <summary>
		/// Constructor. Sets member variables.
		/// </summary>
		/// <param name="keyRange">Key range</param>
		/// <param name="url">URL of the PageBox hosting the Web service</param>
		/// <param name="wsArchive">Name of the archive containing the Web service</param>
		/// <param name="wsName">Web service name</param>
		public PolarisKey(string keyRange, string url, string wsArchive, string wsName) 
		{
			SoapFormatter sf = new SoapFormatter();
			ASCIIEncoding ae = new ASCIIEncoding();
			MemoryStream ms = new MemoryStream(ae.GetBytes(keyRange));
			PolarisKR pkr = (PolarisKR)sf.Deserialize(ms);
			userLow = pkr.userLow;
			userHigh = pkr.userHigh;
			location = pkr.location;
			this.url = computeUrl(url, wsArchive, wsName);
		}

		/// <summary>
		/// Default constructor.
		/// </summary>
		public PolarisKey() 
		{
		}

		/// <summary>
		/// Compute the Web service URL from the PageBox URL, archive name and WebService URL.
		/// Assumption: Web service in a directory named after the archive name under the PageBox URL.
		/// </summary>
		/// <param name="pbUrl">URL of the PageBox hosting the Web service</param>
		/// <param name="wsArchive">Name of the archive containing the Web service</param>
		/// <param name="wsName">Web service name</param>
		/// <returns>Web service URL</returns>
		private string computeUrl(string pbUrl, string wsArchive, string wsName) 
		{
			string[] flds = wsArchive.Split(new char[] { '.' });
			if (!wsName.EndsWith(".asmx"))
				wsName += ".asmx";
			string uri = flds[0] + "/" + wsName;
			return "http://" + pbUrl.Replace("Deploy.asmx", uri);
		}
	}
}
